#!/bin/bash
# binife может читать, записывать/добавлять, удалять значения.

show_help(){
    echo  $"Используйте:""
    $(basename "$0") get" $"inifile section key""
    $(basename "$0") set" $"inifile section key value""
    $(basename "$0") del" $"inifile section key""

"$"Например:""
    $(basename "$0") set .config/myapp.conf behavior killallhumans true

"$"Все аргументы являются обязательными.
Раздел может быть пустым (в качестве аргумента используйте пустые кавычки).
Кавычки удаляются при действии «get», если оно присутствует.

Особенности обработки формата INI:
     Разделы заключаются в '[]'.
     Ключи и значения разделяются знаком «=».
     Ведущие и конечные пробелы вокруг ключей и значений отбрасываются.
     Повторяющиеся пробелы внутри значений без кавычек сжимаются при чтении.
     Значения, содержащие пробелы, записываются в кавычках.
     Ключи, содержащие пробелы, записываются без кавычек.
     Из всех повторяющихся элементов выполняются действия только в первую очередь.
    "
}

echo_error(){
    echo "$@" >&2
}

echo_debug(){
    [ -n "$DEBUG" ] && [ "$DEBUG" -ge "1" ] >/dev/null 2>/dev/null && echo "$@"
}

trim_spaces() {
    local TRIMVAR=$@
    TRIMVAR="${TRIMVAR#"${TRIMVAR%%[![:space:]]*}"}"
    TRIMVAR="${TRIMVAR%"${TRIMVAR##*[![:space:]]}"}"
    echo -n "$TRIMVAR"
}

trim_quotes() {
    local TRIMVAR=$@
    TRIMVAR="${TRIMVAR#[\'\"]}"
    TRIMVAR="${TRIMVAR%[\'\"]}"
    echo -n "$TRIMVAR"
}


trim_single_quotes() {
    local TRIMVAR=$@
    TRIMVAR="${TRIMVAR#[\']}"
    TRIMVAR="${TRIMVAR%[\']}"
    echo -n "$TRIMVAR"
}

trim_double_quotes() {
    local TRIMVAR=$@
    TRIMVAR="${TRIMVAR#[\"]}"
    TRIMVAR="${TRIMVAR%[\"]}"
    echo -n "$TRIMVAR"
}

trim_all_quotes() {
    local TRIMVAR=$@
    TRIMVAR="${TRIMVAR#"${TRIMVAR%%[!\'\"]*}"}"
    TRIMVAR="${TRIMVAR%"${TRIMVAR##*[!\'\"]}"}"
    echo -n "$TRIMVAR"
}

trim_square_brackets(){
    local TRIMVAR=$@
    TRIMVAR="${TRIMVAR#\[}"
    TRIMVAR="${TRIMVAR%\]}"
    echo -n "$TRIMVAR"
}

ini_get(){
    # depends on trim_*, echo_error functions

    local INIFILE="$1"
    local SECTION="$2"
    local KEY="$3"
    
    [ ! -r "$INIFILE" ] && { echo_error $"Can not read file" "\"$INIFILE\"" ; return 1 ; }
    [ -z "$KEY" ] && { echo_error $"No ini key given as argument 3" ; return 1 ; }
    cat "$INIFILE" | while IFS='' read LINE
    do
	# trim spaces around data
	LINE="$(trim_spaces "$LINE")"
	
	# ignore comments
	[ "${LINE:0:1}" = ";" -o "${LINE:0:1}" = "#" ] && continue

	# get section
	[ "${LINE:0:1}" = "[" ] && local DATA_SECTION="$(trim_spaces "$(trim_square_brackets "$LINE")")" && continue

	# ignore line without delimiter
	[ "$LINE" = "${LINE//=/0}" ] && continue

	# get key
	local DATA_KEY="$(trim_spaces "${LINE%%=*}")"

	# if section an key match, get to the data
	if [ "$DATA_SECTION" = "$SECTION" ] && [ "$DATA_KEY" = "$KEY" ]
	then
	    # get value
	    local DATA_VALUE="$(trim_spaces "${LINE#*=}")"

	    # handle paired quotes
	    if [ "${DATA_VALUE:0:1}" = "'" -o "${DATA_VALUE:0:1}" = "\"" ] && [ "${DATA_VALUE:0:1}" = "${DATA_VALUE:(-1):1}" ]
	    then
		# unquote
		local DATA_VALUE="$(trim_quotes "$DATA_VALUE")"
	    else
		#shrink whitespaces if value is unquoted
		local DATA_VALUE="${DATA_VALUE//[[:space:]]/ }"
		until [ "$DATA_VALUE" = "${DATA_VALUE//  / }" ]
		do
		    local DATA_VALUE="${DATA_VALUE//  / }"
		done
	    fi
	    echo "$DATA_VALUE"
	fi
    done
}

ini_set(){
    # depends on trim_*, echo_error functions

    local INIFILE="$1"
    local SECTION="$2"
    local KEY="$3"
    local VALUE="$4"

    # if inifile does not exist, create with newline
    [ ! -e "$INIFILE" ] && { echo_error > "$INIFILE" || echo $"Can not create file" \"$INIFILE\" ; return 1 ; }
    
    # can we read/write the file?
    [ ! -r "$INIFILE" ] && { echo_error $"Can not read file" "\"$INIFILE\"" ; return 1 ; }
    [ ! -w "$INIFILE" ] && { echo_error $"Can not write to file" "\"$INIFILE\"" ; return 1 ; }

    [ -z "$KEY" ] && { echo_error $"No ini key given as argument 3" ; return 1 ; }

    # if value contains whitespaces and is unquoted, quote it
    if [ "$VALUE" != "${VALUE//[[:space:]]/0}" ]
    then
	[ "${VALUE:0:1}" = "'" -o "${VALUE:0:1}" = "\"" ] && [ "${VALUE:0:1}" = "${VALUE:(-1):1}" ] || local VALUE="\"$VALUE\""
    fi

    local REPLACED=false

    local DATA="$(cat "$INIFILE")"
    echo "$DATA" | {
	while IFS='' read ORIG_LINE
	do
	    # trim spaces around data, keep raw line intact
	    local LINE="$(trim_spaces "$ORIG_LINE")"
    
	    # just echo line if comment
	    [ "${LINE:0:1}" = ";" -o "${LINE:0:1}" = "#" ] && { echo "$ORIG_LINE" ; continue ; }
    
	    # if line is a section
	    if [ "${LINE:0:1}" = "[" ] && [ "${LINE:(-1):1}" = "]" ]
	    then
		# if in previous section there were no matches, add value now
		[ "$DATA_SECTION" = "$SECTION" ] && [ "$REPLACED" != "true" ] && { echo "$KEY=$VALUE" ; REPLACED=true ; }
		# get section
		local DATA_SECTION="$(trim_spaces "$(trim_square_brackets "$LINE")")"
		# return section line
		echo "$ORIG_LINE"
		continue
	    fi
    
	    # just echo line without delimiter
	    [ "$LINE" = "${LINE//=/0}" ] && { echo "$ORIG_LINE" ; continue ; }
    
	    # get key
	    local DATA_KEY="$(trim_spaces "${LINE%%=*}")"
    
	    # if section and key match and no replacemet took place,
	    # echo original line with updated value (saving anything we can, but old value), otherwise echo original line
	    if [ "$DATA_SECTION" = "$SECTION" ] && [ "$DATA_KEY" = "$KEY" ] &&  [ "$REPLACED" != "true" ]
	    then
		echo "${ORIG_LINE%%=*}=$VALUE"
		local REPLACED=true
	    else
		echo "${ORIG_LINE}"
	    fi
	done 
	# add section and key if not found inside above cycle
	if [ "$REPLACED" != "true" ]
	then
	    [ -n "$SECTION" ] && [ "$DATA_SECTION" != "$SECTION" ] && echo "[$SECTION]"
	    echo -e "$KEY=$VALUE"
	fi
    } > "$INIFILE"
}

ini_del(){
    # depends on trim_*, echo_error functions

    local INIFILE="$1"
    local SECTION="$2"
    local KEY="$3"

    # if inifile does not exist do nothing
    [ ! -e "$INIFILE" ] && { echo_error $"File does not exist:" \"$INIFILE\" ; return 1 ; }
    
    # can we read/write the file?
    [ ! -r "$INIFILE" ] && { echo_error $"Can not read file" "\"$INIFILE\"" ; return 1 ; }
    [ ! -w "$INIFILE" ] && { echo_error $"Can not write to file" "\"$INIFILE\"" ; return 1 ; }

    [ -z "$KEY" ] && { echo_error $"No ini key given as argument 3" ; return 1 ; }

    local REMOVED=false

    local DATA="$(cat "$INIFILE")"
    echo "$DATA" | {
	while IFS='' read ORIG_LINE
	do
	    # trim spaces around data, keep raw line intact
	    local LINE="$(trim_spaces "$ORIG_LINE")"
    
	    # just echo line if comment
	    [ "${LINE:0:1}" = ";" -o "${LINE:0:1}" = "#" ] && { echo "$ORIG_LINE" ; continue ; }
    
	    # if line is a section
	    if [ "${LINE:0:1}" = "[" ] && [ "${LINE:(-1):1}" = "]" ]
	    then
		# get section
		local DATA_SECTION="$(trim_spaces "$(trim_square_brackets "$LINE")")"
		# return section line
		echo "$ORIG_LINE"
		continue
	    fi
    
	    # just echo line without delimiter
	    [ "$LINE" = "${LINE//=/0}" ] && { echo "$ORIG_LINE" ; continue ; }
    
	    # get key
	    local DATA_KEY="$(trim_spaces "${LINE%%=*}")"
    
	    # if section and key match and nothing got removed so far, do not return line
	    # echo original line with updated value (saving anything we can, but old value), otherwise echo original line
	    if [ "$DATA_SECTION" = "$SECTION" ] && [ "$DATA_KEY" = "$KEY" ] &&  [ "$REMOVED" != "true" ]
	    then
		local REMOVED=true
		true
	    else
		echo "${ORIG_LINE}"
	    fi
	done 
    } > "$INIFILE"
}

ACTION="$1"
shift
echo "$ACTION" | grep -qi 'help\|\-h' && { show_help ; exit 0 ; }

if [ "$ACTION" = "get" ] && [ "$#" = "3" ]
then
    ini_get "$@"
elif [ "$ACTION" = "set" ] && [ "$#" = "4" ]
then
    ini_set "$@"
elif [ "$ACTION" = "del" ] && [ "$#" = "3" ]
then
    ini_del "$@"
else
    echo_error "Неверные аргументы для действия $ACTION"
    show_help >&2
    exit 1
fi
